{ File    : $URL: file:///D:/SVN/svnDelphi/trunk/BS5/VT7AExt/src/VT7AExt.dpr $
  Author  : $Author: Deniz Oezmen $
  Changed : $Date: 2013-12-29 23:25:54 +0100 (So, 29 Dez 2013) $
  Revision: $Rev: 240 $

  Broken Sword 5 VT7A/VSA extractor.
}
program VT7AExt;

{$APPTYPE CONSOLE}

{$R '..\res\version.res' '..\res\version.rc'}

uses
  ModInfo in 'utils\ModInfo.pas',
  SysUtils,
  StrUtils,
  Classes,
  IniFiles,
  NativeXml,
  ZLib;

const
  Command_None    = $00000000;
  Command_Invalid = $00000001;
  Command_Info    = $00000002;
  Command_Operate = $00000004;

  VT7AHeaderID = 'VT7A';
  VSAHeaderID  = 'AUFS';

  SpeechPath = 'audio/speech/';

  GraphicsLocs: array[0..69] of AnsiString = ('bijou_apt',
    'bijou_apt_dresser_closeup', 'bijou_exterior1', 'gallery',
    'gallery_alarm_closeup', 'gallery_exterior',
    'gallery_exterior_cafe_closeup', 'gallery_henri_closeup',
    'gallery_laine_closeup', 'gallery_machine_closeup_scene_17',
    'gallery_office', 'gallery_office_cctv_closeup',
    'gallery_office_drawer_closeup', 'gallery_office_safe',
    'gallery_office_sofa_closeup', 'gallery_pizza_closeup',
    'gallery_simeon_scene_17', 'hobbs_corpse_closeup',
    'hobbs_portfolio_closeup', 'hobbs_stereo_closeup', 'hobbs_studio',
    'hobbs_studio_night', 'hobbs_upstairs', 'hobbs_upstairs_pinboard',
    'hobbs_yard', 'hobbs_yard_escape', 'hobbs_yard_establishing',
    'hobbs_yard_intro_scene_23', 'hobbs_yard_night',
    'hobbs_yard_night_van_cab_closeup', 'hobbs_yard_van_cab_closeup',
    'hobbs_yard_van_engine_closeup', 'map_paris_closeup', 'nico_apt',
    'nico_apt_cdplayer_closeup', 'nico_apt_exterior',
    'nico_apt_fleur_shop_exterior', 'nico_apt_fleur_shop_interior',
    'nico_apt_hallway', 'nico_apt_neighbour', 'nico_apt_photos_closeup',
    'object_hobbs_sketch', 'object_la_malediccio',
    'object_marques_family_photo', 'object_marques_notes',
    'object_simeons_notes', 'russian_house_cabinet_closeup',
    'russian_house_cigarette_box_closeup', 'russian_house_establishing',
    'russian_house_exterior', 'russian_house_front_door_closeup',
    'russian_house_library', 'russian_house_study',
    'russian_house_study_desk_closeup', 'russian_house_typewriter_closeup',
    'tutorial_touch_1', 'tutorial_touch_2', 'tutorial_touch_3',
    'vera_ashtray_closeup', 'vera_cockroach_closeup', 'vera_security_alley',
    'vera_security_alley_night', 'vera_security_inner_office',
    'vera_security_inner_office_hand_closeup', 'vera_security_neonsign_scene_5',
    'vera_security_outer_office', 'vera_security_outer_office_night',
    'vera_security_shredder_puzzle', 'vera_shredder_closeup',
    'vera_shutter_closeup');

type
  TParamRec = record
    FileSpec   : AnsiString;
    NameSpec   : AnsiString;
    AddNameSpec: AnsiString;
    UseNames   : Boolean;
    GenNames   : Boolean;
    SkipUnknown: Boolean;
    Info       : Boolean;
  end;

  TVT7AHeader = packed record
    ID      : array[0..3] of Char;
    Version : Longword;
    Unknown2: Longword;
    NumFiles: Longword;
  end;

  TVSAHeader = packed record
    ID      : array[0..3] of Char;
    NumFiles: Longword;
  end;

  // generalized header variant
  THeader = TVSAHeader;

  TVT7ADirEntry = packed record
    Hash      : Longint;
    Offset    : Longword;
    UncompSize: Longword;
    CompSize  : Longword;
  end;

  TVT7ADirEntries = packed array of TVT7ADirEntry;

  TVSADirEntry = packed record
    Hash      : Longint;
    Offset    : Longword;
    UncompSize: Longword;
  end;

  TVSADirEntries = packed array of TVSADirEntry;

  // generalized directory variant
  TDirEntry = TVT7ADirEntry;
  TDirEntries = packed array of TDirEntry;

  TBuffer = array of Byte;

var
  ParamRec   : TParamRec;
  Command    : Longword;
  NameTable  : THashedStringList;
  AddNameFile: TextFile;

function OpenInFile(var InFile: file; InName: string): Boolean;
begin
  FileMode := fmShareDenyWrite;
  AssignFile(InFile, InName);

  {$I-}
  Reset(InFile, 1);
  {$I+}
  Result := (IOResult = 0)
end;

function OpenOutFile(var OutFile: file; OutName: string): Boolean;
var
  Dir: AnsiString;
begin
  // first create the directory portion of the output file
  Dir := ExtractFileDir(ExpandFileName(OutName));
  ForceDirectories(Dir);

  // open the output file
  FileMode := fmOpenWrite;
  AssignFile(OutFile, OutName);

  {$I-}
  Rewrite(OutFile, 1);
  {$I+}
  Result := (IOResult = 0)
end;

function ReadHeader(var VT7AFile: file; out Header: THeader): Boolean;
var
  ID        : array[0..3] of Char;
  VT7AHeader: TVT7AHeader;
  VSAHeader : TVSAHeader;
  BytesRead : Longint;
begin
  Result := False;

  BlockRead(VT7AFile, ID, SizeOf(ID), BytesRead);
  if BytesRead <> SizeOf(ID) then
    Exit;

  Seek(VT7AFile, 0);

  if ID = VT7AHeaderID then
  begin
    BlockRead(VT7AFile, VT7AHeader, SizeOf(VT7AHeader), BytesRead);
    Result := (BytesRead = SizeOf(VT7AHeader));

    Header.ID := VT7AHeaderID;
    Header.NumFiles := VT7AHeader.NumFiles
  end
  else if ID = VSAHeaderID then
  begin
    BlockRead(VT7AFile, VSAHeader, SizeOf(VSAHeader), BytesRead);
    Result := (BytesRead = SizeOf(VSAHeader));

    Header.ID := VSAHeaderID;
    Header.NumFiles := VSAHeader.NumFiles
  end
end;

function Decompress(const InBuffer: TBuffer; InOfs, InLen: Longint;
  var OutBuffer: TBuffer; OutOfs, OutLen: Longint): Boolean;
var
  DecompSize: Longint;
  TempBuffer: Pointer;
begin
  // standard ZLib decompression
  DecompressBuf(@InBuffer[InOfs], InLen, OutLen, TempBuffer,
    DecompSize);

  Result := (DecompSize = OutLen);
  if Result then
    Move(TempBuffer^, OutBuffer[OutOfs], DecompSize);

  FreeMem(TempBuffer, DecompSize)
end;

function HashName(const Name: AnsiString): Longint;
var
  i: Longint;
begin
  Result := 0;

  for i := 1 to Length(Name) do
    Result := Result shl 7 + Result shl 1 + Result + Shortint(Name[i])
end;

{function ReverseHash(var Hash: Longint; LastChar: Char): Boolean;
var
  Bit,
  BitIdx,
  TempHash,
  LastHash : Longint;
begin
  LastHash := 0;
  TempHash := Hash;

  TempHash := TempHash - Ord(LastChar);

  for BitIdx := 0 to 31 do
  begin
    Bit := TempHash and (1 shl BitIdx);
    LastHash := LastHash + Bit;
    TempHash := TempHash - Bit - (Bit shl 1) - (Bit shl 7)
  end;

  TempHash := (LastHash shl 7) + (LastHash shl 1) + LastHash + Ord(LastChar);

  Result := (Hash = TempHash);

  if Result then
    Hash := LastHash;
end;}

function ReadNames: Boolean;
var
  Count: Longint;
  Names: TextFile;
  Line,
  Ident: AnsiString;
begin
  Result := False;

  NameTable := THashedStringList.Create;

  AssignFile(Names, ParamRec.NameSpec);
  {$I-}
  Reset(Names);
  {$I+}
  if IOResult <> 0 then
  begin
    WriteLn('error: cannot read name file');
    Exit
  end;

  // count lines
  Count := 0;
  while not Eof(Names) do
  begin
    Inc(Count);
    ReadLn(Names)
  end;

  // prepare table and read names
  NameTable.Capacity := Count;

  Reset(Names);
  while not Eof(Names) do
  begin
    ReadLn(Names, Line);

    // Special case for speech files: Only the pure file name (without path or
    // extension) is used for hashing. (Or is this rather a special case for
    // VSA archives?)
    if LeftStr(Line, Length(SpeechPath)) = SpeechPath then
      Ident := MidStr(Line, Length(SpeechPath) + 1, Length(Line) -
        Length(SpeechPath) - Length(ExtractFileExt(Line)))
    else
      Ident := Line;
    Ident := IntToHex(HashName(Ident), 8);

    NameTable.Add(Ident + '=' + Line)
  end;

  CloseFile(Names);

  Result := True
end;

procedure GenerateXMLNames(const XMLName: AnsiString;
  out AddNames: TStringList);

  procedure ScanNode(const Node: TXmlNode; const RelDirName: AnsiString;
    const AddNames: TStringList);
  var
    i, p       : Longint;
    Images     : TStringList;
    GraphicsLoc,
    XMLName2   : AnsiString;
    XMLDoc2    : TNativeXML;
  begin
    // look for certain special nodes
    if Node.Name = 'hotspot-poly' then
      AddNames.Add(RelDirName + Node.AttributeValueByName['name'] + '.xml')
    else if Node.Name = 'poly-walk-area' then
      AddNames.Add(RelDirName + Node.AttributeValueByName['name'] + '.xml')
    else if (Node.Name = 'object') and (Node.NodeCount = 1) then
      AddNames.Add(RelDirName + Node.Nodes[0].Value + '.xml')
    else if Node.Name = 'router' then
      AddNames.Add(RelDirName + Node.AttributeValueByName['grid'])
    else if (Node.Name = 'sfx-group-item') and (Node.NodeCount = 1) then
      AddNames.Add('audio/sfx/' + Node.Nodes[0].Value + '.ogg')
    else if (Node.Name = 'location') then
    begin
      Images := TStringList.Create;

      for i := 0 to Node.NodeCount - 1 do
        if Node.Nodes[i].Name = 'background' then
          Images.Add(Node.Nodes[i].Nodes[0].Value)
        else if Node.Nodes[i].Name = 'foreground' then
          Images.Add(Node.Nodes[i].Nodes[0].Value)
        else if Node.Nodes[i].Name = 'graphics-location' then
          GraphicsLoc := Node.Nodes[i].Nodes[0].Value;

      if (GraphicsLoc <> '') then
      begin
        for i := 0 to Images.Count - 1 do
        begin
          AddNames.Add(Format('graphics/1280x720/location/%s/%s.xml',
            [GraphicsLoc, Images[i]]));
          AddNames.Add(Format('graphics/1280x720/location/%s/%s.tex',
            [GraphicsLoc, Images[i]]))
        end;

        AddNames.Add(Format('graphics/1280x720/location/%s/lightmask.tex',
          [GraphicsLoc]));
        AddNames.Add(Format('graphics/1280x720/location/%s/lightmask2.tex',
          [GraphicsLoc]));
        AddNames.Add(Format('graphics/1280x720/location/%s/shadowmap.sdw',
          [GraphicsLoc]));
        AddNames.Add(Format('graphics/1280x720/location/%s/shadowmap2.sdw',
          [GraphicsLoc]))
      end;

      Images.Free
    end
    else if Node.Name = 'graphic' then
    begin
      // try to get a matching location XML
      i := Pos('/', XMLName);
      p := -1;

      if i > 0 then
        p := PosEx('/', XMLName, i + 1);

      if p > 0 then
        XMLName2 := Copy(XMLName, 1, p) + Copy(XMLName, i + 1, p - i - 1) +
          '.xml';

      GraphicsLoc := '';
      if XMLName2 <> '' then
      begin
        // load XML file
        XMLDoc2 := TNativeXml.Create(nil);
        XMLDoc2.LoadFromFile(XMLName2);

        if XMLDoc2.Root.Name = 'location' then
          for i := 0 to XMLDoc2.Root.NodeCount - 1 do
            if XMLDoc2.Root.Nodes[i].Name = 'graphics-location' then
            begin
              GraphicsLoc := XMLDoc2.Root.Nodes[i].Value;
              Break
            end;

        if GraphicsLoc <> '' then
        begin
          AddNames.Add(Format('graphics/1280x720/location/%s/animation/%s.xml',
            [GraphicsLoc, Node.AttributeValueByName['name']]));
          AddNames.Add(Format('graphics/1280x720/location/%s/animation/%s.tex',
            [GraphicsLoc, Node.AttributeValueByName['name']]));
          AddNames.Add(Format('graphics/1280x720/location/%s/animation/%s.str' +
            'eam', [GraphicsLoc, Node.AttributeValueByName['name']]))
        end
      end
    end;

    // scan children recursively
    for i := 0 to Node.NodeCount - 1 do
      if Node.Nodes[i].ElementType = xeElement then
        ScanNode(Node.Nodes[i], RelDirName, AddNames)
  end;

var
  XMLDoc    : TNativeXml;
  RelDirName: AnsiString;
begin
  // load XML file
  XMLDoc := TNativeXml.Create(nil);
  XMLDoc.LoadFromFile(XMLName);

  // prepare string list for unique new names
  AddNames := TStringList.Create;
  AddNames.Sorted := True;
  AddNames.Duplicates := dupIgnore;

  // get only the relative path name from the XML file name
  RelDirName := ExtractRelativePath(IncludeTrailingPathDelimiter(
    GetCurrentDir), ExtractFilePath(ExpandFileName(XMLName)));
  RelDirName := AnsiReplaceStr(RelDirName, '\', '/');

  // parse all nodes
  ScanNode(XMLDoc.Root, RelDirName, AddNames);

  XMLDoc.Free
end;

procedure GenerateJOEYNames(const JOEYName: AnsiString;
  out AddNames: TStringList);
const
  NameChars: set of Char = [' ', '!', '#', '$', '%', '&', '''', '(', ')', '+',
                            ',', '-', '.', '0', '1', '2',  '3', '4', '5', '6',
                            '7', '8', '9', ';', '=', '@',  'A', 'B', 'C', 'D',
                            'E', 'F', 'G', 'H', 'I', 'J',  'K', 'L', 'M', 'N',
                            'O', 'P', 'Q', 'R', 'S', 'T',  'U', 'V', 'W', 'X',
                            'Y', 'Z', '[', ']', '^', '_',  '`', 'a', 'b', 'c',
                            'd', 'e', 'f', 'g', 'h', 'i',  'j', 'k', 'l', 'm',
                            'n', 'o', 'p', 'q', 'r', 's',  't', 'u', 'v', 'w',
                            'x', 'y', 'z', '{', '}', '~'];
var
  JOEYFile   : file;
  Buffer     : TBuffer;
  BufIdx,
  i, p       : Longint;
  Size       : Word;
  Name,
  XMLName,
  RelDirName,
  GraphicsLoc: AnsiString;
  XMLDoc     : TNativeXml;
begin
  // prepare string list for unique new names
  AddNames := TStringList.Create;
  AddNames.Sorted := True;
  AddNames.Duplicates := dupIgnore;

  // get only the relative path name from the JOEY file name
  RelDirName := ExtractRelativePath(IncludeTrailingPathDelimiter(
    GetCurrentDir), ExtractFilePath(ExpandFileName(JOEYName)));
  RelDirName := AnsiReplaceStr(RelDirName, '\', '/');

  // try to get a matching location XML
  i := Pos('/', JOEYName);
  p := -1;
  
  if i > 0 then
    p := PosEx('/', JOEYName, i + 1);

  if p > 0 then
    XMLName := Copy(JOEYName, 1, p) + Copy(JOEYName, i + 1, p - i - 1) + '.xml';

  GraphicsLoc := '';
  if XMLName <> '' then
  begin
    // load XML file
    XMLDoc := TNativeXml.Create(nil);
    XMLDoc.LoadFromFile(XMLName);

    if XMLDoc.Root.Name = 'location' then
      for i := 0 to XMLDoc.Root.NodeCount - 1 do
        if XMLDoc.Root.Nodes[i].Name = 'graphics-location' then
        begin
          GraphicsLoc := XMLDoc.Root.Nodes[i].Value;
          Break
        end
  end;

  // load JOEY file
  if not OpenInFile(JOEYFile, JOEYName) then
  begin
    WriteLn(Format('    warning: cannot open %s to generate names',
      [JOEYName]));
    Exit
  end;

  // load the entire file into memory
  {$I-}
  SetLength(Buffer, FileSize(JOEYFile));
  BlockRead(JOEYFile, Buffer[0], Length(Buffer));
  {$I+}
  if IOResult = 0 then
  begin
    // This is some kind of brute force approach since we're not really parsing
    // the JOEY file (, though we maybe could ...): Every string (that could be
    // a potential sound file name) is prepended by a word indicating the size
    // of the string including the size descriptor itself. So we basically just
    // scan the whole file byte by byte looking for size/string pairs matching
    // this description.

    BufIdx := 0;
    while BufIdx <= High(Buffer) - 1 do
    begin
      // get the possible size value
      Move(Buffer[BufIdx], Size, SizeOf(Size));

      // check whether this is actually a feasible size
      if (Size < 2) or (BufIdx + Size > High(Buffer)) then
      begin
        Inc(BufIdx);
        Continue
      end;

      // the size is basically okay, so now check the character validity (except
      // the last one, see below)
      Name := '';
      for i := 2 to Longint(Size) - 2 do
      begin
        if not (Char(Buffer[BufIdx + i]) in NameChars) then
          Break
        else if i = Size - SizeOf(Size) then
        begin
          // we might have a candidate, which has to pass one last check
          SetLength(Name, Size - SizeOf(Size));
          Move(Buffer[BufIdx + SizeOf(Size)], Name[1], Length(Name))
        end
      end;

      if Name = '' then
      begin
        Inc(BufIdx);
        Continue
      end;

      // the character data up to now was okay, so we're doing a special check
      // on the last character: This one might be zero, but does not have to be.
      if Name[Length(Name)] = #000 then
        SetLength(Name, Length(Name) - 1)
      else if not (Char(Name[Length(Name)]) in NameChars) then
      begin
        Inc(BufIdx);
        Continue
      end;

      // we have a candidate name ...
      AddNames.Add(SpeechPath + Name + '.ogg');
      AddNames.Add('audio/sfx/' + Name + '.ogg');
      AddNames.Add('audio/music/' + Name + '.ogg');
      AddNames.Add('graphics/1280x720/icons/' + Name + '.tex');
      AddNames.Add('graphics/1280x720/icons/' + Name + '.xml');
      AddNames.Add('graphics/1280x720/mega/george_' + Name + '.mega');
      AddNames.Add('graphics/1280x720/mega/nico_' + Name + '.mega');

      if LeftStr(Name, 5) = 'look-' then
        AddNames.Add(RelDirName + RightStr(Name, Length(Name) - 5) + '.xml');
      if RightStr(Name, 5) = '-look' then
        AddNames.Add(RelDirName + LeftStr(Name, Length(Name) - 5) + '.xml');
      if LeftStr(Name, 5) = 'look_' then
        AddNames.Add(RelDirName + RightStr(Name, Length(Name) - 5) + '.xml');
      if RightStr(Name, 5) = '_look' then
        AddNames.Add(RelDirName + LeftStr(Name, Length(Name) - 5) + '.xml');
      if LeftStr(Name, 9) = 'interact-' then
        AddNames.Add(RelDirName + RightStr(Name, Length(Name) - 9) + '.xml');
      if RightStr(Name, 9) = '-interact' then
        AddNames.Add(RelDirName + LeftStr(Name, Length(Name) - 9) + '.xml');
      if LeftStr(Name, 9) = 'interact_' then
        AddNames.Add(RelDirName + RightStr(Name, Length(Name) - 9) + '.xml');
      if RightStr(Name, 9) = '_interact' then
        AddNames.Add(RelDirName + LeftStr(Name, Length(Name) - 9) + '.xml');
      if LeftStr(Name, 6) = 'logic-' then
        AddNames.Add(RelDirName + RightStr(Name, Length(Name) - 6) + '.xml');
      if RightStr(Name, 6) = '-logic' then
        AddNames.Add(RelDirName + LeftStr(Name, Length(Name) - 6) + '.xml');
      if LeftStr(Name, 6) = 'logic_' then
        AddNames.Add(RelDirName + RightStr(Name, Length(Name) - 6) + '.xml');
      if RightStr(Name, 6) = '_logic' then
        AddNames.Add(RelDirName + LeftStr(Name, Length(Name) - 6) + '.xml');

      if GraphicsLoc <> '' then
      begin
        AddNames.Add(Format('graphics/1280x720/location/%s/animation/%s.stream',
          [GraphicsLoc, Name]));
        AddNames.Add(Format('graphics/1280x720/location/%s/animation/%s.tex',
          [GraphicsLoc, Name]));
        AddNames.Add(Format('graphics/1280x720/location/%s/animation/%s.xml',
          [GraphicsLoc, Name]))
      end;

      if RightStr(RelDirName, 8) = 'globals/' then
        for i := 0 to High(GraphicsLocs) do
        begin
          AddNames.Add(Format('graphics/1280x720/location/%s/animation/%s.str' +
            'eam', [GraphicsLocs[i], Name]));
          AddNames.Add(Format('graphics/1280x720/location/%s/animation/%s.tex',
            [GraphicsLocs[i], Name]));
          AddNames.Add(Format('graphics/1280x720/location/%s/animation/%s.xml',
            [GraphicsLocs[i], Name]))
        end;

      Inc(BufIdx, Size)
    end
  end
  else
  begin
    WriteLn(Format('    warning: could not read from %s to generate names',
      [JOEYName]));
  end;

  CloseFile(JOEYFile)
end;

procedure ExtractFiles(var VT7AFile: file; const VT7AName: AnsiString;
  const Header: THeader);
var
  i, j          : Longint;
  VT7ADirEntries: TVT7ADirEntries;
  VSADirEntries : TVSADirEntries;
  DirEntries    : TDirEntries;
  OutName       : AnsiString;
  Buffer,
  UncompBuf     : TBuffer;
  OutFile       : file;
  AddNames      : TStringList;
begin
  if Header.NumFiles = 0 then
  begin
    WriteLn('  archive is empty');
    Exit
  end;

  // read directory table
  if Header.ID = VT7AHeaderID then
  begin
     SetLength(VT7ADirEntries, Header.NumFiles);
     {$I-}
     BlockRead(VT7AFile, VT7ADirEntries[0],
       Length(VT7ADirEntries) * SizeOf(TVT7ADirEntry));
     {$I+}

     SetLength(DirEntries, Header.NumFiles);
     Move(VT7ADirEntries[0], DirEntries[0],
       SizeOf(TVT7ADirEntry) * Length(VT7ADirEntries))
  end
  else if Header.ID = VSAHeaderID then
  begin
     SetLength(VSADirEntries, Header.NumFiles);
     {$I-}
     BlockRead(VT7AFile, VSADirEntries[0],
       Length(VSADirEntries) * SizeOf(TVSADirEntry));
     {$I+}

     SetLength(DirEntries, Header.NumFiles);
     for i := 0 to High(VSADirEntries) do
     begin
       DirEntries[i].Hash := VSADirEntries[i].Hash;
       DirEntries[i].Offset := VSADirEntries[i].Offset;
       DirEntries[i].UncompSize := VSADirEntries[i].UncompSize;
       DirEntries[i].CompSize := 0
     end
  end
  else
  begin
    WriteLn('  internal error: unsupported header format');
    SetLength(DirEntries, 0);
    Exit
  end;

  if IOResult <> 0 then
  begin
    WriteLn('  error reading structures');
    SetLength(DirEntries, 0);
    Exit
  end;

  // process all files
  for i := 0 to High(DirEntries) do
    with DirEntries[i] do
    begin
      // determine a name for this file
      OutName := '';
      if Assigned(NameTable) then
        OutName := NameTable.Values[IntToHex(Hash, 8)];

      if OutName = '' then
      begin
        if ParamRec.SkipUnknown then
          Continue
        else
          OutName := Format('%s_%.8x.dat', [VT7AName, Hash])
      end;

      WriteLn('  ', IntToHex(Hash, 8) + ' ' + OutName);

      if CompSize = 0 then
        // uncompressed file
        SetLength(Buffer, UncompSize)
      else
        // compressed file
        SetLength(Buffer, CompSize);

      // read file data
      {$I-}
      Seek(VT7AFile, Offset);
      BlockRead(VT7AFile, Buffer[0], Length(Buffer));
      {$I+}
      if IOResult <> 0 then
      begin
        WriteLn('    read error');
        Continue
      end;

      // optionally decompress a compressed file
      if CompSize > 0 then
      begin
        SetLength(UncompBuf, UncompSize);
        
        if Decompress(Buffer, 0, Length(Buffer), UncompBuf, 0,
          Length(UncompBuf)) then
          Buffer := UncompBuf
        else
          WriteLn('    decompression error')
      end;

      // open output file and write final data
      if not OpenOutFile(OutFile, OutName) then
      begin
        WriteLn('    error opening output file');
        Continue
      end;

      {$I-}
      BlockWrite(OutFile, Buffer[0], Length(Buffer));
      {$I+}
      if IOResult <> 0 then
        WriteLn('    write error');

      CloseFile(OutFile);

      // if we know this file to be an XML file, try go extract additional file
      // names from it
      if ParamRec.GenNames and (ExtractFileExt(OutName) = '.xml') then
      begin
        GenerateXMLNames(OutName, AddNames);

        for j := 0 to AddNames.Count - 1 do
          WriteLn(AddNameFile, AddNames[j]);

        AddNames.Free
      end
      // if we know this file to be a JOEY file, try go extract additional file
      // names from it
      else if ParamRec.GenNames and (ExtractFileExt(OutName) = '.joey') then
      begin
        GenerateJOEYNames(OutName, AddNames);

        // we can currently only get *possible* file names from JOEY files, so
        // we just take whatever we can get ...
        for j := 0 to AddNames.Count - 1 do
          WriteLn(AddNameFile, AddNames[j]);

        AddNames.Free
      end
    end;

  SetLength(Buffer, 0);
  SetLength(UncompBuf, 0);
  SetLength(DirEntries, 0)
end;

procedure ScanFiles;
var
  FileName,
  Dir,
  VT7AName  : string;
  VT7AFile  : file;
  Header    : THeader;
  SearchRec : TSearchRec;
begin
  // get the directory information
  Dir := ExtractFilePath(ParamRec.FileSpec);

  // search for all files matching the file specification
  if FindFirst(ParamRec.FileSpec, faAnyFile - faDirectory,
    SearchRec) = 0 then
  begin
    repeat
      FileName := SearchRec.Name;
      VT7AName := Dir + FileName;

      WriteLn(VT7AName);

      if OpenInFile(VT7AFile, VT7AName) then
      begin
        if ReadHeader(VT7AFile, Header) then
          ExtractFiles(VT7AFile, FileName, Header)
        else
          WriteLn('  not a valid VT7A or VSA file')
      end
      else
        WriteLn('  not accessible')
    until FindNext(SearchRec) <> 0
  end
  else
    WriteLn(Format('No matches for "%s"', [ParamRec.FileSpec]));

  SysUtils.FindClose(SearchRec)
end;

function ProcessParams: Longword;
var
  i         : Longint;
  Param,
  ParamLower: AnsiString;
begin
  Result := Command_None;

  if ParamCount = 0 then
  begin
    Result := Result or Command_Invalid;
    Exit
  end;

  with ParamRec do
  begin
    FileSpec    := '';
    NameSpec    := '';
    AddNameSpec := '';
    UseNames    := False;
    GenNames    := False;
    SkipUnknown := False;
    Info        := False;

    for i := 1 to ParamCount do
    begin
      Param := ParamStr(i);
      ParamLower := AnsiLowerCase(Param);
      if Copy(ParamLower, 1, 7) = '/names=' then
      begin
        UseNames := True;
        if NameSpec <> '' then
          Result := Result or Command_Invalid;

        NameSpec := Copy(Param, 8, Length(Param) - 7)
      end
      else if Copy(ParamLower, 1, 15) = '/generatenames=' then
      begin
        GenNames := True;
        if AddNameSpec <> '' then
          Result := Result or Command_Invalid;

        AddNameSpec := Copy(Param, 16, Length(Param) - 15)
      end
      else if ParamLower = '/skipunknown' then
        SkipUnknown := True
      else if ParamLower = '/info' then
        Info := True
      else if FileSpec = '' then
        FileSpec := Param
      else
        Result := Result or Command_Invalid
    end;

    if Info then
      Result := Result or Command_Info;

    if GenNames and not UseNames then
      Result := Result or Command_Invalid;

    if SkipUnknown and not UseNames then
      Result := Result or Command_Invalid;

    if UseNames and (FileSpec = '') then
      Result := Result or Command_Invalid;

    if (FileSpec <> '') and ((Result and Command_Invalid) = 0) then
      Result := Result or Command_Operate
  end
end;

begin
  Command := ProcessParams;

  if (Command and Command_Info) <> 0 then
    WriteLn(GetModuleFileVersionString);
  if (Command and Command_Invalid) <> 0 then
    WriteLn('Usage: VT7AExt <VT7AFiles/VSAFiles> [/names=<NameFile> [/generat' +
      'enames=<AdditionalNameFile>] [/skipunknown]] [/info]');
  if (Command and Command_Operate) <> 0 then
  begin
    // read name table if available
    if ParamRec.UseNames then
      if not ReadNames then
        Exit;

    if ParamRec.GenNames then
    begin
      {$I-}
      AssignFile(AddNameFile, ParamRec.AddNameSpec);
      Rewrite(AddNameFile);
      {$I+}
      if IOResult <> 0 then
      begin
        WriteLn(Format('error writing to %s', [ParamRec.AddNameSpec]));
        if Assigned(NameTable) then
          NameTable.Free;

        Exit
      end
    end;

    ScanFiles;

    if Assigned(NameTable) then
      NameTable.Free;
    {$I-}
    CloseFile(AddNameFile);
    {$I+}
  end
end.
